/* readSGP input */

use io;

function input() {
    usage = "\nUsage: localsolver SGP080519.lsp "
    + "inFileName=inputFile solFileName=outputFile [lsTimeLimit=timeLimit]\n";

    if (inFileName == nil || solFileName == nil) throw usage;

    readInputSGP();

    computeTimeTraversalSmall();
    computeTimeTraversalLarge();
    computeDepotTravelTimeSmall();
    computeDepotTravelTimeLarge();

    writeOutput();
}

function model() {
	
	nodeSequences[k in 1..nSnowcats] <- list(nbNodes);  

    constraint disjoint[k in 1..nSnowcats](nodeSequences[k]);

    nJobs = jobID[nbNodes-1];
	
    for [j in 1..nJobs]{
		
    	for [i in 0..(nbNodes-1)]{
			if (j == jobID[i]){
				jobWidth[j] = nodeWidth[i];
    			jobType[j] = nodeType[i];
    			jobLength[j] = nodeLength[i];
    		}
    	}
		
    	traversalSmall[j] <- sum[k in 1..nSmallSnowcats][i in 0..nbNodes : jobID[i] == j](contains(nodeSequences[k], i));
		
    	if (jobWidth[j] ==1){
    		traversalLarge[j] <- sum[k in 1+nSmallSnowcats..nSnowcats][i in 0..nbNodes : jobID[i] == j](contains(nodeSequences[k], i));
    		constraint traversalLarge[j] == 0;
    	}
		
		traversalLarge[j] <- sum[k in 1+nSmallSnowcats..nSnowcats][i in 0..nbNodes : jobID[i] == j](contains(nodeSequences[k], i));
		
		jobComplete[j] <- traversalSmall[j] + 2 * traversalLarge[j] >= jobWidth[j]; 
		
		if (jobType[j] == 2){		
			constraint jobComplete[j] == 1;
		}
		
    }
	
	for [k in 1..nSnowcats] {
		local sequence <- nodeSequences[k];
		local c <- count(sequence);

		// A snowcat is used if it visits at least one node
		snowcatUsed[k] <- c > 0;

		
		if (k <= nSmallSnowcats){
			routeDistance[k] <- sum(1..c-1, 
				i => nodeMatrixDist[sequence[i-1]][sequence[i]])
				+ (snowcatUsed[k] ? 
				(depotDistance[k][sequence[0]] + depotDistance[k][sequence[c-1]]) :
				0);
			constraint routeDistance[k] <= rangeSnowcats[k];

			/*totalDistance[k] <- sum(1..c-1, i => nodeMatrixDist[sequence[i-1]][sequence[i]]);  //range constraint
			constraint totalDistance[k] <= rangeSnowcats[k];*/

			endTime[k] <- array(0..c-1, (i, prev) => max(startTW[sequence[i]],
								i == 0 ?
								depotTravelTimeSmall[k][sequence[0]] :
								prev + nodeMatrixTimeSmall[sequence[i-1]][sequence[i]])
							+ nodeDuration[sequence[i]] );		
							
			homeLateness[k] <- snowcatUsed[k] ?
				max(0, endTime[k][c-1] + depotDistance[k][sequence[c-1]] - availableWorkmins) :
				0;
		}

		if (k > nSmallSnowcats){
			routeDistance[k] <- sum(1..c-1, 
				i => nodeMatrixDistLarge[sequence[i-1]][sequence[i]])
				+ (snowcatUsed[k] ? 
				(depotDistanceLarge[k][sequence[0]] + depotDistanceLarge[k][sequence[c-1]]) :
				0);
			constraint routeDistance[k] <= rangeSnowcats[k];

			/*totalDistance[k] <- sum(1..c-1, i => nodeMatrixDistLarge[sequence[i-1]][sequence[i]]);  //range constraint
			constraint totalDistance[k] <= rangeSnowcats[k];*/

			endTime[k] <- array(0..c-1, (i, prev) => max(startTW[sequence[i]],
								i == 0 ?
								depotTravelTimeLarge[k][sequence[0]] :
								prev + nodeMatrixTimeLarge[sequence[i-1]][sequence[i]])
							+ nodeDuration[sequence[i]] );	
							
			homeLateness[k] <- snowcatUsed[k] ?
				max(0, endTime[k][c-1] + depotDistanceLarge[k][sequence[c-1]] - availableWorkmins) :
				0;
		}

		lateness[k] <- homeLateness[k] + sum(0..c-1,
					i => max(0, endTime[k][i] - endTW[sequence[i]]));
	}

	//Total lateness, must be 0 for a solution to be valid
	totalLateness <- sum[k in 1..nSnowcats](lateness[k]);

	constraint totalLateness == 0;

	completed <- sum[j in 1..nJobs](jobComplete[j]);
	distance <- sum[k in 1..nSnowcats](routeDistance[k]);

	nonPriDist <- sum[j in 1..nJobs : jobType[j] != 1](jobLength[j]*jobComplete[j]);

	priDist <- priArcProfit * sum[j in 1..nJobs : jobType[j] == 1](jobLength[j]*jobComplete[j]);

	totalTime <- sum[k in 1..nSnowcats](routeDistance[k] / (operatingSpeed * (1000/60)));

	distanceCovered <- nonPriDist + priDist;

	objective <- nonPriDist + priDist - 0.1 * totalTime;

	maximize objective;
}

/* Parameterizes the solver. */
function param() { 
    if (lsTimeLimit == nil) lsTimeLimit = 10;
    if (lsNbThreads == nil) lsNbThreads = 2;
    if (lsTimeBetweenDisplays == nil) lsTimeBetweenDisplays = 1;
    getSolutionStatus();
}

function output()
{
	for[k in 1..nSnowcats]
	{
		io.stdout.println("Snowcat " + k + ": routeDistance = " + routeDistance[k].value);
		local tour <- nodeSequences[k];
		
		prev = nil;
		for[i in tour.value]{
			if (prev != nil)
			{
				io.stdout.println(nodeMatrixDist[prev][i] + " (" + nodeLength[prev] + ")");
			}
			io.stdout.print(nodeStart[i] + "->" + nodeEnd[i] + ": ");
			prev = i;
		}
		io.stdout.println("\n");
		io.stdout.println("endTime for Snowcat" + k + ":" + endTime[k].value);
		io.stdout.println("\n");
	}
	/*
	io.stdout.println("\n");
	
	io.stdout.println();
	for[j in 1..nJobs]{
		io.stdout.println("Job " + j + ": width = " + jobWidth[j] + " type = " + jobType[j]);
		io.stdout.println(" completed : " + jobComplete[j].value);
		io.stdout.print(" traversed large = " + traversalLarge[j].value);
		io.stdout.println(" traversed small = " + traversalSmall[j].value);
		io.stdout.println(" totalLateness = " + totalLateness.value);
	}*/
	io.stdout.println();
	io.stdout.println("nodeNodes: " + nodeNodes);
	io.stdout.println("nbNodes: " + nbNodes);
	io.stdout.println("nIntersections: " + nIntersections);
	io.stdout.println("Total unique distance covered: " + distanceCovered.value);
	io.stdout.println();


}

function readInputSGP() {
	local inFile = io.openRead(inFileName);
	inFile.readString();
	nodeNodes = inFile.readInt();
	inFile.readString();
	nArcs = inFile.readInt();
	inFile.readString();
	nIntersections = inFile.readInt();
	inFile.readString();
	nSnowcats = inFile.readInt();
	inFile.readString();
	nSmallSnowcats = inFile.readInt();
	inFile.readString();
	nLargeSnowcats = inFile.readInt();
	inFile.readString();
	operatingSpeed = inFile.readInt();
	inFile.readString();
	availableWorkmins = inFile.readInt();
	inFile.readString();
	priArcProfit = inFile.readInt();

	// Skip 2 lines, not 3 because the first empty line is consumed by the 
	// previous call to readInt().
	skipLines(inFile, 2);

	// Job Data
    i = 0;
	while(true) {
		inLine = inFile.readln();
		line = inLine.replace("\t", " ").split();
		if (count(line) == 0) break;
		if (count(line) != 10) throw "Wrong file format";
		jobID[i] = toInt(line[0]);
		nodeID[i] = toInt(line[1]);
		nodeStart[i] = toInt(line[2]);
		nodeEnd[i] = toInt(line[3]);
		nodeLength[i] = toInt(line[4]);
		nodeDuration[i] = toDouble(line[5]);
		nodeType[i] = toInt(line[6]);
		nodeWidth[i] = toInt(line[7]);
		startTW[i] = toInt(line[8]);
		endTW[i] = toInt(line[9]);
		i = i + 1;
	}
	nbNodes = i;
	
	skipLines(inFile, 1);
	
	nodeMatrixDist[0..nodeNodes-1][0..nodeNodes-1] = inFile.readInt();

	skipLines(inFile, 1);
	
    nodeMatrixDistLarge[0..nodeNodes-1][0..nodeNodes-1] = inFile.readInt();
	
	skipLines(inFile, 1);
	
	rangeSnowcats[1..nSnowcats] = inFile.readInt();

	skipLines(inFile, 1);

	// TODO
	
	Depots = nSnowcats;

	//depotdistance for the small snowcats
	depotDistance[1..Depots][0..nodeNodes-1] = inFile.readInt();
	
	skipLines(inFile, 1);
	
	//depotdistance for the large snowcats
	depotDistanceLarge[1..Depots][0..nodeNodes-1] = inFile.readInt();
	
	if(!inFile.eof()) throw "End of file expected";
	inFile.close();
}

function writeOutput() {
	local outFile= io.openWrite(solFileName);
	outFile.println(nbNodes);
	outFile.println();
	outFile.println("Width of nodes: ");
	for[i in 0..nbNodes-1] outFile.println(nodeWidth[i]);
	/*outFile.println("Objective value: ", objective.value);*/

}

function computeTimeTraversalSmall() {
	for[p in 0..nbNodes-1][q in 0..nbNodes-1]{
		nodeMatrixTimeSmall[p][q] = nodeMatrixDist[p][q] / (operatingSpeed * (1000/60));
	}
}

function computeTimeTraversalLarge() {
	for[p in 0..nbNodes-1][q in 0..nbNodes-1]{
		nodeMatrixTimeLarge[p][q] = nodeMatrixDistLarge[p][q] / (operatingSpeed * (1000/60));
	}
}

function computeDepotTravelTimeSmall() {
	depotTravelTimeSmall[i in 1..Depots][j in 0..nodeNodes-1] = depotDistance[i][j] / (operatingSpeed * (1000/60));
}

function computeDepotTravelTimeLarge() {
	depotTravelTimeLarge[i in 1..Depots][j in 0..nodeNodes-1] = depotDistanceLarge[i][j] / (operatingSpeed * (1000/60));
}

function skipLines(inFile, nbLines) {
    if (nbLines < 1) return;
    local dump = inFile.readln();
    for[i in 2..nbLines] dump = inFile.readln();
}


